#![allow(dead_code)]

use std::{
    collections::BTreeMap,
    fmt::{Debug, Display},
};

use intern::{InternPoolWrapper as InternPool, PointerFlags, StructFlags, AMD64_POINTER_TYPE_INFO};
use num_bigint::BigInt;

use crate::{
    ast::FloatingType, comptime::ComptimeNumber, lexer::SourceLocation,
    symbol_table::syms2::DeclKind, tokens::Token,
};

pub mod debug;
pub mod intern;
pub mod ir;
pub mod parser;
pub mod tag;
pub mod visitor;

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum Tag {
    /// pseudo tag, contains a range from a..b into extra of all files.
    Root,
    /// `data` is a range from a..b into extra of all global nodes.
    File,
    /// `data` is an intern to a name, and an index into extra of [index: return_type, index: ParameterList]
    FunctionProto,
    /// `data` is an intern to a name, and an intern to the function type
    FunctionProtoInterned,
    /// `data` is an index to a FunctionProto and an index to a Block
    FunctionDecl,
    /// `data` is a range from a..b into extra of indices to parameters
    ParameterList,
    /// `data` is an index to a type, and an intern to a name
    Parameter,
    /// `data` is range from a..b into `extra` of indices to statements
    Block,
    /// `data` is range from a..b into `extra` of indices to statements, where the last one is an expression
    BlockTrailingExpr,
    /// `data` is an index to a type, and an intern to a value
    Constant,
    /// `data` is an index to an expression
    ExprStmt,
    /// `data` is none
    ReturnStmt,
    /// `data` is an index to an expr
    ReturnExprStmt,
    /// `data` is a range from a..b into `extra` of `[name: intern, type: index]`
    VarDecl,
    /// `data` is a range from a..b into `extra` of `[name: intern, type: index]`
    MutVarDecl,
    /// `data` is a range from a..b into `extra` of `[name: intern, expr: index, type?: index]`
    VarDeclAssignment,
    /// `data` is a range from a..b into `extra` of `[name: intern, expr: index, type?: index]`
    MutVarDeclAssignment,
    /// `data` is an intern to a name, and an offset into `extra` of `[type: index, expr: index]`
    GlobalDecl,
    /// `data` is an intern to a name, and an offset into extra of `[flags, type0 ,..., typeN ,name0 ,..., nameN]`
    StructDecl,
    /// `data` is an intern to a name, and an intern to the type of the struct
    StructDeclInterned,
    /// `data` is an index to a type, and an intern to a name
    FieldDecl,
    /// `data` is an index to a Parameter, VarDecl, GlobalDecl or FunctionDecl, and an opaque DeclKind
    DeclRef,
    /// `data` is an inlined key into the symbol table (scope: index, name: intern)
    DeclRefUnresolved,
    /// `data` is an intern of a type
    InternedType,
    /// `data` is an index to a StructDecl
    TypeDeclRef,
    /// `data` is an inlined key into the symbol table (scope: index, name: intern)
    TypeDeclRefUnresolved,
    /// `data` is an index to a Type and u32 PointerFlags (extra offset)
    PointerType,
    /// `data` is an index to a length expression, and an underlying pointer type
    ArrayType,
    /// `data` is an index to an expr and an index to an ArgumentList
    CallExpr,
    /// `data` is an index to an expr and an intern to a field name
    FieldAccess,
    /// `data` is a range from a..b into extra of indices to arguments
    ArgumentList,
    /// `data` is an index to an expression
    Argument,
    /// `data` is an index to an expression, and an intern to a name
    NamedArgument,
    /// `data` is an index to lhs, and an index to the type
    ExplicitCast,
    /// `data` is a single index to an expr
    Deref,
    AddressOf,
    Not,
    Negate,
    PlaceToValueConversion,
    ValueToPlaceConversion,
    /// data is two indices for `lhs` and `rhs`
    Or,
    And,
    BitOr,
    BitXOr,
    BitAnd,
    Eq,
    NEq,
    Lt,
    Gt,
    Le,
    Ge,
    Shl,
    Shr,
    Add,
    Sub,
    Mul,
    Div,
    Rem,
    Assign,
    SubscriptExpr,
    IfExpr,
    /// `data` is an index to an expression and an index into extra for [if, else]
    IfElseExpr,
    // TODO:
    /// `data` is a ParseError
    Error,
    /// placeholder tag for reserved indices/nodes, `data` is none
    Undefined,
}

impl Tag {
    fn is_type(&self) -> bool {
        match self {
            Tag::TypeDeclRef
            | Tag::TypeDeclRefUnresolved
            | Tag::PointerType
            | Tag::InternedType
            | Tag::ArrayType
            | Tag::StructDeclInterned
            | Tag::StructDecl => true,
            _ => false,
        }
    }
    fn is_expr(&self) -> bool {
        match self {
            Tag::Constant
            | Tag::DeclRef
            | Tag::Deref
            | Tag::CallExpr
            | Tag::AddressOf
            | Tag::Not
            | Tag::Negate
            | Tag::Or
            | Tag::And
            | Tag::BitOr
            | Tag::BitXOr
            | Tag::BitAnd
            | Tag::Eq
            | Tag::NEq
            | Tag::Lt
            | Tag::Gt
            | Tag::Le
            | Tag::Ge
            | Tag::Shl
            | Tag::Shr
            | Tag::Add
            | Tag::Sub
            | Tag::Mul
            | Tag::Div
            | Tag::Rem
            | Tag::SubscriptExpr
            | Tag::IfExpr
            | Tag::IfElseExpr
            | Tag::PlaceToValueConversion
            | Tag::ValueToPlaceConversion
            | Tag::Block
            | &Tag::BlockTrailingExpr => true,
            _ => false,
        }
    }
}

#[derive(Debug, Clone, Copy, thiserror::Error, PartialEq, Eq)]
pub enum ParseError {
    #[error("Unexpected end of token iter.")]
    UnexpectedEndOfTokens,
    #[error("Expected Token {0}.")]
    ExpectedToken(Token),
    #[error("Expected Token {0}, but other token was found.")]
    ExpectedTokenNotFound(Token),
    #[error("Expected either a function declaration or a global variable.")]
    UnexpectedTokenAtFileScope,
    #[error("Expected Ident.")]
    ExpectedIdent,
    #[error("Integral types may not be wider than 65535 bits.")]
    IntegralTypeTooWide,
    #[error("Expected typename.")]
    ExpectedTypeName,
    #[error("Dummy Message.")]
    ExpectedFunctionPrototype,
    #[error("Dummy Message.")]
    ExpectedPrimaryExpression,
    #[error("Dummy Message.")]
    ExpectedConstantLiteral,
    #[error("Dummy Message.")]
    ExpectedExpression,
    #[error("Dummy Message.")]
    ExpectedPostfixExpression,
    #[error("Dummy Message.")]
    ExpectedPrefixExpression,
    #[error("Dummy Message.")]
    ExpectedArgumentList,
    #[error("Dummy Message.")]
    ExpectedStatement,
    #[error("Dummy Message.")]
    UnmatchedParens(u32),
    #[error("Dummy Message.")]
    ExpectedTypeDeclaration,
    #[error("Dummy Message.")]
    UnexpectedTypeAttributes,
    #[error("Dummy Message.")]
    UnmatchedSquareBracket(u32),
    #[error("Dummy Message.")]
    ExpectedEndOfBlock,
    #[error("Dummy Message.")]
    UnmatchedBrace(u32),
    #[error("Dummy Message.")]
    UnmatchedDelimiter(u32),
    #[error("Error in child node {0:?}.")]
    ErrorNode(Index),
}

#[repr(packed)]
#[derive(Clone, Copy)]
struct Node {
    /// defines the type of the node in the tree
    tag: Tag,
    data: Data,
}

#[derive(Clone, Copy)]
pub union Data {
    none: (),
    error: ParseError,
    index: Index,
    two_indices: (Index, Index),
    range: (Index, Index),
    extra_range: (u32, u32),
    intern: intern::Index,
    index_intern: (Index, intern::Index),
    two_interns: (intern::Index, intern::Index),
    intern_and_extra_offset: (intern::Index, u32),
    index_and_extra_offset: (Index, u32),
    index_and_opaque: (Index, u32),
}

impl Default for Tag {
    fn default() -> Self {
        Tag::Undefined
    }
}
impl Default for SourceLocation {
    fn default() -> Self {
        Self::invalid()
    }
}

impl Default for Data {
    fn default() -> Self {
        Self::none()
    }
}

#[derive(Debug)]
#[allow(dead_code)]
enum ExpandedData {
    None,
    Error(ParseError),
    Index(Index),
    TwoIndices(Index, Index),
    Range(Index, Index),
    ExtraRange(usize, usize),
    Intern(intern::Index),
    IndexIntern(Index, intern::Index),
    TwoInterns(intern::Index, intern::Index),
    InternAndExtraOffset(intern::Index, usize),
    IndexAndExtraOffset(Index, usize),
    IndexAndOpaque(Index, u32),
}

impl ExpandedData {
    fn from_none() -> Self {
        Self::None
    }
    fn from_error(data: Data) -> Self {
        Self::Error(data.as_error())
    }
    fn from_index(data: Data) -> Self {
        Self::Index(data.as_index())
    }

    fn from_two_indices(data: Data) -> Self {
        let data = data.as_two_indices();
        Self::TwoIndices(data.0, data.1)
    }
    fn from_range(data: Data) -> Self {
        let data = data.as_index_range();
        Self::Range(data.0, data.1)
    }
    fn from_extra_range(data: Data) -> Self {
        let data = data.as_extra_range();
        Self::ExtraRange(data.0, data.1)
    }
    fn from_intern(data: Data) -> Self {
        let data = data.as_intern();
        Self::Intern(data)
    }
    fn from_index_intern(data: Data) -> Self {
        let data = data.as_index_intern();
        Self::IndexIntern(data.0, data.1)
    }
    fn from_two_interns(data: Data) -> Self {
        let data = data.as_two_interns();
        Self::TwoInterns(data.0, data.1)
    }
    fn from_intern_and_extra_offset(data: Data) -> Self {
        let data = data.as_intern_and_extra_offset();
        Self::InternAndExtraOffset(data.0, data.1)
    }
    fn from_index_and_extra_offset(data: Data) -> Self {
        let data = data.as_index_and_extra_offset();
        Self::IndexAndExtraOffset(data.0, data.1)
    }
    fn from_index_and_opaque(data: Data) -> Self {
        let data = data.as_index_and_opaque();
        Self::IndexAndExtraOffset(data.0, data.1)
    }
}

impl From<(Tag, Data)> for ExpandedData {
    fn from((tag, data): (Tag, Data)) -> Self {
        match tag {
            Tag::FunctionProto => Self::from_index_and_extra_offset(data),
            Tag::ParameterList => Self::from_extra_range(data),
            Tag::Root => Self::from_extra_range(data),
            Tag::File => Self::from_extra_range(data),
            Tag::ArgumentList
            | Tag::VarDecl
            | Tag::MutVarDecl
            | Tag::VarDeclAssignment
            | Tag::MutVarDeclAssignment
            | Tag::BlockTrailingExpr
            | Tag::Block => Self::from_extra_range(data),
            Tag::FieldDecl | Tag::Constant | Tag::Parameter => Self::from_index_intern(data),
            Tag::Or
            | Tag::And
            | Tag::BitOr
            | Tag::BitXOr
            | Tag::BitAnd
            | Tag::Eq
            | Tag::NEq
            | Tag::Lt
            | Tag::Gt
            | Tag::Le
            | Tag::Ge
            | Tag::Shl
            | Tag::Shr
            | Tag::Add
            | Tag::Sub
            | Tag::Mul
            | Tag::Div
            | Tag::Rem
            | Tag::Assign
            | Tag::IfExpr
            | Tag::SubscriptExpr
            | Tag::CallExpr
            | Tag::ArrayType
            | Tag::FunctionDecl => Self::from_two_indices(data),
            Tag::StructDeclInterned | Tag::FunctionProtoInterned => Self::from_two_interns(data),
            Tag::ReturnExprStmt
            | Tag::DeclRef
            | Tag::TypeDeclRef
            | Tag::Argument
            | Tag::Deref
            | Tag::AddressOf
            | Tag::Not
            | Tag::Negate
            | Tag::PlaceToValueConversion
            | Tag::ValueToPlaceConversion
            | Tag::ExprStmt => Self::from_index(data),
            Tag::FieldAccess
            | Tag::DeclRefUnresolved
            | Tag::TypeDeclRefUnresolved
            | Tag::NamedArgument
            | Tag::ExplicitCast => Self::from_index_intern(data),
            Tag::GlobalDecl => Self::from_intern_and_extra_offset(data),
            Tag::InternedType | Tag::StructDecl => Self::from_intern(data),
            Tag::PointerType | Tag::IfElseExpr => Self::from_index_and_extra_offset(data),
            Tag::Error => Self::from_error(data),
            Tag::ReturnStmt | Tag::Undefined => Self::from_none(),
        }
    }
}

impl Data {
    fn as_error(self) -> ParseError {
        unsafe { self.error }
    }
    fn as_index(self) -> Index {
        unsafe { self.index }
    }
    fn as_two_indices(self) -> (Index, Index) {
        unsafe { self.two_indices }
    }
    fn as_index_range(self) -> (Index, Index) {
        unsafe { self.range }
    }
    fn as_extra_range(self) -> (usize, usize) {
        let (a, b) = unsafe { self.extra_range };
        (a as usize, b as usize)
    }
    fn as_intern(self) -> intern::Index {
        unsafe { self.intern }
    }

    fn as_two_interns(self) -> (intern::Index, intern::Index) {
        unsafe { self.two_interns }
    }

    fn as_index_intern(self) -> (Index, intern::Index) {
        unsafe { self.index_intern }
    }
    fn as_index_and_extra_offset(self) -> (Index, usize) {
        let (i, e) = unsafe { self.index_and_extra_offset };
        (i, e as usize)
    }
    fn as_intern_and_extra_offset(self) -> (intern::Index, usize) {
        let (i, e) = unsafe { self.intern_and_extra_offset };
        (i, e as usize)
    }
    fn as_index_and_opaque(self) -> (Index, usize) {
        let (i, e) = unsafe { self.index_and_opaque };
        (i, e as usize)
    }
}

impl Data {
    fn none() -> Self {
        Self { none: () }
    }
    fn error(error: ParseError) -> Self {
        Self { error }
    }
    fn index(index: Index) -> Self {
        Self { index }
    }
    fn two_indices(a: Index, b: Index) -> Self {
        Self {
            two_indices: (a, b),
        }
    }
    fn two_interns(a: intern::Index, b: intern::Index) -> Self {
        Self {
            two_interns: (a, b),
        }
    }
    fn range_of_indices(a: Index, b: Index) -> Self {
        Self { range: (a, b) }
    }
    fn extra_range(a: u32, b: u32) -> Self {
        Self {
            extra_range: (a, b),
        }
    }
    fn intern(intern: intern::Index) -> Self {
        Self { intern }
    }
    fn index_and_intern(index: Index, intern: intern::Index) -> Self {
        Self {
            index_intern: (index, intern),
        }
    }
    fn intern_and_extra_offset(intern: intern::Index, offset: u32) -> Self {
        Self {
            intern_and_extra_offset: (intern, offset),
        }
    }
    fn index_and_extra_offset(index: Index, offset: u32) -> Self {
        Self {
            index_and_extra_offset: (index, offset),
        }
    }
    fn index_and_opaque(index: Index, value: u32) -> Self {
        Self {
            index_and_opaque: (index, value),
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum PlaceOrValue {
    Value(index::Index),
    Place(index::Index),
}

impl PlaceOrValue {
    fn into_index(self) -> index::Index {
        match self {
            PlaceOrValue::Value(index) => index,
            PlaceOrValue::Place(index) => index,
        }
    }
    fn is_value(&self) -> bool {
        matches!(self, &Self::Value(_))
    }
    fn is_place(&self) -> bool {
        !self.is_value()
    }
    fn into_value(self) -> PlaceOrValue {
        Self::Value(self.into_index())
    }
    fn into_place(self) -> PlaceOrValue {
        Self::Place(self.into_index())
    }
    fn eq_discriminant(&self, other: &Self) -> bool {
        core::mem::discriminant(self).eq(&core::mem::discriminant(other))
    }
    fn with_index(self, index: Index) -> PlaceOrValue {
        match self {
            PlaceOrValue::Value(_) => Self::Value(index),
            PlaceOrValue::Place(_) => Self::Place(index),
        }
    }
    fn map_value<F>(self, f: F) -> Self
    where
        F: FnOnce(Index) -> Self,
    {
        match self {
            PlaceOrValue::Value(index) => f(index),
            _ => self,
        }
    }
    fn map_place<F>(self, f: F) -> Self
    where
        F: FnOnce(Index) -> Self,
    {
        match self {
            PlaceOrValue::Place(index) => f(index),
            _ => self,
        }
    }
}

impl From<index::Index> for PlaceOrValue {
    fn from(value: index::Index) -> Self {
        Self::Value(value)
    }
}

mod index {
    use std::num::NonZero;

    #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
    #[repr(u8)]
    pub enum Kind {
        Other = 0,
        Function,
        GlobalDecl,
        Error,
        File,
    }

    impl TryFrom<u8> for Kind {
        type Error = ();

        fn try_from(value: u8) -> Result<Self, ()> {
            match value {
                0 => Ok(Self::Other),
                1 => Ok(Self::Function),
                2 => Ok(Self::GlobalDecl),
                3 => Ok(Self::Error),
                4 => Ok(Self::File),
                _ => Err(()),
            }
        }
    }

    /// Index type that has 28 bits of index and 4 bits of kind.
    #[repr(transparent)]
    #[derive(Clone, Copy, Hash, PartialEq, Eq, PartialOrd, Ord)]
    pub struct Index(NonZero<u32>);

    impl core::fmt::Debug for Index {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            f.debug_struct("Index")
                .field("kind", &self.kind())
                .field("index", &self.index_u32())
                .finish()
        }
    }

    impl core::fmt::Display for Index {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            let label = match self.kind() {
                Kind::Other => "",
                Kind::Function => "funcs.",
                Kind::GlobalDecl => "gdecls.",
                Kind::Error => "errors.",
                Kind::File => "files.",
            };

            write!(f, "%{label}{}", self.index_u32())
        }
    }

    impl Index {
        pub fn new(kind: Kind, idx: u32) -> Option<Self> {
            let inner = NonZero::new((kind as u32) << 28 | idx & ((1 << 28) - 1));

            inner.map(Self)
        }

        pub fn from_slice(slice: &[u32]) -> &[Option<Self>] {
            unsafe { core::mem::transmute(slice) }
        }
        pub unsafe fn from_slice_unchecked(slice: &[u32]) -> &[Self] {
            unsafe { core::mem::transmute(slice) }
        }

        pub fn kind(&self) -> Kind {
            Kind::try_from(((self.0.get() >> 28) & 0b1111) as u8).unwrap()
        }
        pub fn index_u32(&self) -> u32 {
            self.0.get() & ((1 << 28) - 1)
        }
        pub fn index_usize(&self) -> usize {
            self.index_u32() as usize
        }
        pub fn from_u32(inner: u32) -> Option<Self> {
            NonZero::new(inner).map(Self)
        }
        pub unsafe fn from_u32_unchecked(inner: u32) -> Self {
            Self(NonZero::new_unchecked(inner))
        }
        pub fn as_u32(self) -> u32 {
            self.0.get()
        }
        pub fn into_u32(self) -> u32 {
            self.0.get()
        }
    }
}

pub use index::{Index, Kind};

#[derive(Debug, Clone)]
struct AstTables<T> {
    functions: Vec<T>,
    global_decls: Vec<T>,
    errors: Vec<T>,
    files: Vec<T>,
    other: Vec<T>,
}

impl<T> core::ops::Index<index::Index> for AstTables<T> {
    type Output = T;

    fn index(&self, index: index::Index) -> &Self::Output {
        &self.table_ref(index.kind())[index.index_usize()]
    }
}

impl<T> core::ops::IndexMut<index::Index> for AstTables<T> {
    fn index_mut(&mut self, index: index::Index) -> &mut Self::Output {
        &mut self.table_mut(index.kind())[index.index_usize()]
    }
}

impl<T> AstTables<T> {
    fn new() -> Self {
        Self {
            functions: Vec::new(),
            global_decls: Vec::new(),
            errors: Vec::new(),
            files: Vec::new(),
            other: Vec::new(),
        }
    }

    fn get(&self, index: index::Index) -> Option<&T> {
        self.table_ref(index.kind()).get(index.index_usize())
    }
    unsafe fn get_unchecked(&self, index: index::Index) -> &T {
        self.table_ref(index.kind())
            .get_unchecked(index.index_usize())
    }

    fn init() -> Self
    where
        T: Default,
    {
        Self {
            functions: Vec::new(),
            global_decls: Vec::new(),
            errors: Vec::new(),
            files: Vec::new(),
            other: vec![T::default()],
        }
    }

    fn push(&mut self, kind: index::Kind, t: T) -> index::Index {
        let table = self.table_mut(kind);
        let i = table.len();
        table.push(t);

        index::Index::new(kind, i as u32).unwrap()
    }

    fn reserve(&mut self, kind: index::Kind) -> index::Index
    where
        T: Default,
    {
        self.push(kind, T::default())
    }

    fn table_mut(&mut self, kind: index::Kind) -> &mut Vec<T> {
        match kind {
            index::Kind::Other => &mut self.other,
            index::Kind::Function => &mut self.functions,
            index::Kind::GlobalDecl => &mut self.global_decls,
            index::Kind::Error => &mut self.errors,
            index::Kind::File => &mut self.files,
        }
    }

    fn table_ref(&self, kind: index::Kind) -> &Vec<T> {
        match kind {
            index::Kind::Other => &self.other,
            index::Kind::Function => &self.functions,
            index::Kind::GlobalDecl => &self.global_decls,
            index::Kind::Error => &self.errors,
            index::Kind::File => &self.files,
        }
    }

    fn table_len(&self, kind: index::Kind) -> usize {
        self.table_ref(kind).len()
    }

    fn iter(&self) -> impl Iterator<Item = &T> {
        self.errors
            .iter()
            .chain(self.files.iter())
            .chain(self.functions.iter())
            .chain(self.global_decls.iter())
            .chain(self.other.iter())
    }
}

pub struct Ast {
    tags: AstTables<Tag>,
    datas: AstTables<Data>,
    source_locs: AstTables<SourceLocation>,
    extra: Vec<u32>,
}

impl Debug for Ast {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Ast")
            .field_with("nodes", |f| {
                let mut list = f.debug_list();
                struct LocDisplay(SourceLocation);
                impl Debug for LocDisplay {
                    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                        write!(f, "({})", self.0)
                    }
                }
                let entries = self
                    .tags
                    .iter()
                    .cloned()
                    .zip(self.datas.iter().cloned())
                    .zip(self.source_locs.iter().cloned())
                    .enumerate()
                    .map(|(i, ((tag, data), loc))| {
                        (i, tag, ExpandedData::from((tag, data)), LocDisplay(loc))
                    });
                list.entries(entries).finish()
            })
            .field("extra", &self.extra)
            .finish()
    }
}

impl Ast {
    fn new() -> Ast {
        let mut tags = AstTables::new();
        let mut datas = AstTables::new();
        let mut source_locs = AstTables::new();

        tags.other.push(Tag::Root);
        datas.other.push(Data::extra_range(0, 0));
        source_locs.other.push(SourceLocation::new(0, 0));

        Self {
            tags,
            datas,
            extra: vec![],
            source_locs,
        }
    }

    fn reserve_node_other(&mut self) -> Index {
        self.reserve_node(Kind::Other)
    }

    fn reserve_node(&mut self, kind: index::Kind) -> Index {
        let i = self.tags.reserve(kind);
        self.datas.reserve(kind);
        self.source_locs.reserve(kind);

        i
    }

    fn get_loc(&self, index: Index) -> SourceLocation {
        self.source_locs[index]
    }

    fn push_error(&mut self, error: ParseError, loc: SourceLocation) -> Index {
        let i = self.reserve_node(Kind::Error);
        self.set_tag_data_source_loc(i, Tag::Error, Data::error(error), loc);

        i
    }

    fn set_file<I: IntoIterator<Item = Index>>(&mut self, i: Index, decls: I, loc: SourceLocation) {
        let (extra_start, extra_end) = self.extend_extra_by_indices(decls);
        self.set_tag_data_source_loc(i, Tag::File, Data::extra_range(extra_start, extra_end), loc);
    }

    fn push_file<I: IntoIterator<Item = Index>>(&mut self, decls: I, loc: SourceLocation) -> Index {
        let i = self.reserve_node(Kind::File);
        self.set_file(i, decls, loc);
        i
    }

    fn set_root<I: IntoIterator<Item = Index>>(&mut self, decls: I) {
        let (extra_start, extra_end) = self.extend_extra_by_indices(decls);
        self.tags.other[0] = Tag::Root;
        self.datas.other[0] = Data::extra_range(extra_start, extra_end);
    }

    fn get_root_file_indices<'a>(&'a self) -> impl Iterator<Item = Index> + 'a {
        let (a, b) = self.datas.other[0].as_extra_range();
        self.extra[a..b]
            .iter()
            .cloned()
            .map(|i| Index::from_u32(i).unwrap())
    }

    fn push_global_decl(
        &mut self,
        ident: intern::Index,
        ty: Index,
        expr: Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node(Kind::GlobalDecl);
        let (extra_start, _) = self.extend_extra([ty.into_u32(), expr.into_u32()]);
        self.set_tag_data_source_loc(
            i,
            Tag::GlobalDecl,
            Data::intern_and_extra_offset(ident, extra_start),
            loc,
        );

        i
    }

    fn set_fn_decl(&mut self, i: Index, proto: Index, body: Index, loc: SourceLocation) {
        self.set_tag_data_source_loc(i, Tag::FunctionDecl, Data::two_indices(proto, body), loc);
    }

    fn push_fn_decl(&mut self, proto: Index, body: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node(Kind::Function);
        self.set_fn_decl(i, proto, body, loc);
        i
    }

    fn push_ret(&mut self, expr: Option<Index>, loc: SourceLocation) -> Index {
        let i = self.reserve_node(Kind::Other);
        match expr {
            Some(expr) => {
                self.set_tag_data_source_loc(i, Tag::ReturnExprStmt, Data::index(expr), loc)
            }
            None => self.set_tag_data_source_loc(i, Tag::ReturnStmt, Data::none(), loc),
        }

        i
    }

    fn push_var_decl(
        &mut self,
        is_let: bool,
        name: intern::Index,
        ty: Option<Index>,
        assignment: Option<Index>,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();

        let start = self.extra.len() as u32;
        self.extra.push(name.into_u32());
        _ = self.extend_extra(assignment.map(|i| i.into_u32()));
        _ = self.extend_extra(ty.map(|i| i.into_u32()));
        let end = self.extra.len() as u32;

        let tag = match (is_let, assignment.is_some()) {
            (true, false) => Tag::VarDecl,
            (true, true) => Tag::VarDeclAssignment,
            (false, false) => Tag::MutVarDecl,
            (false, true) => Tag::MutVarDeclAssignment,
        };

        self.set_tag_data_source_loc(i, tag, Data::extra_range(start, end), loc);

        i
    }

    fn push_fn_proto(
        &mut self,
        ident: intern::Index,
        return_type: Index,
        parameter_list: Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        let (extra_start, _) =
            self.extend_extra([return_type.into_u32(), parameter_list.into_u32()]);
        self.set_tag_data_source_loc(
            i,
            Tag::FunctionProto,
            Data::intern_and_extra_offset(ident, extra_start),
            loc,
        );

        i
    }

    fn set_block<I: IntoIterator<Item = Index>>(
        &mut self,
        i: Index,
        statements: I,
        trailing: Option<Index>,
        loc: SourceLocation,
    ) {
        let (extra_start, extra_end) =
            self.extend_extra_by_indices(statements.into_iter().chain(trailing.into_iter()));
        if trailing.is_some() {
            self.set_tag_data_source_loc(
                i,
                Tag::BlockTrailingExpr,
                Data::extra_range(extra_start, extra_end),
                loc,
            );
        } else {
            self.set_tag_data_source_loc(
                i,
                Tag::Block,
                Data::extra_range(extra_start, extra_end),
                loc,
            );
        }
    }

    fn push_block<I: IntoIterator<Item = Index>>(
        &mut self,
        statements: I,
        trailing: Option<Index>,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        self.set_block(i, statements, trailing, loc);
        i
    }

    fn push_parameter_list<I: IntoIterator<Item = Index>>(
        &mut self,
        parameters: I,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        let (extra_start, extra_end) = self.extend_extra_by_indices(parameters);
        self.set_tag_data_source_loc(
            i,
            Tag::ParameterList,
            Data::extra_range(extra_start, extra_end),
            loc,
        );

        i
    }

    fn push_argument(&mut self, expr: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::Argument, Data::index(expr), loc);

        i
    }

    fn push_named_argument(
        &mut self,
        name: intern::Index,
        expr: Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(
            i,
            Tag::NamedArgument,
            Data::index_and_intern(expr, name),
            loc,
        );

        i
    }

    fn push_parameter(&mut self, name: intern::Index, ty: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::Parameter, Data::index_and_intern(ty, name), loc);

        i
    }

    fn push_argument_list<I: IntoIterator<Item = Index>>(
        &mut self,
        args: I,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        let (extra_start, extra_end) = self.extend_extra_by_indices(args);
        self.set_tag_data_source_loc(
            i,
            Tag::ArgumentList,
            Data::extra_range(extra_start, extra_end),
            loc,
        );

        i
    }

    fn push_unary(&mut self, tag: Tag, lhs: Index, loc: SourceLocation) -> PlaceOrValue {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, tag, Data::index(lhs), loc);

        match tag {
            Tag::Deref => PlaceOrValue::Place(i),
            _ => PlaceOrValue::Value(i),
        }
    }

    /// converts from a place expression to a value expression.
    fn push_place_to_value_conversion(&mut self, index: Index) -> PlaceOrValue {
        let loc = self.get_loc(index);
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::PlaceToValueConversion, Data::index(index), loc);

        PlaceOrValue::Value(i)
    }

    /// converts from a value expression to a place expression.
    fn push_value_to_place_conversion(&mut self, index: Index) -> PlaceOrValue {
        let loc = self.get_loc(index);
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::ValueToPlaceConversion, Data::index(index), loc);

        PlaceOrValue::Place(i)
    }

    /// converts the expression to a value expression, if it isn't one already.
    fn convert_to_value_expr(&mut self, lrvalue: PlaceOrValue) -> Index {
        match lrvalue {
            PlaceOrValue::Value(index) => index,
            PlaceOrValue::Place(index) => self.push_place_to_value_conversion(index).into_index(),
        }
    }

    /// converts the expression to a place expression, if it isn't one already.
    fn convert_to_place_expr(&mut self, lrvalue: PlaceOrValue) -> Index {
        match lrvalue {
            PlaceOrValue::Place(index) => index,
            PlaceOrValue::Value(index) => self.push_value_to_place_conversion(index).into_index(),
        }
    }

    fn push_binary(&mut self, tag: Tag, lhs: Index, rhs: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, tag, Data::two_indices(lhs, rhs), loc);

        i
    }

    fn push_assign(&mut self, lhs: Index, rhs: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::Assign, Data::two_indices(lhs, rhs), loc);

        i
    }

    fn push_cast(&mut self, lhs: Index, ty: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::ExplicitCast, Data::two_indices(lhs, ty), loc);

        i
    }

    fn push_if(&mut self, cond: Index, body: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::IfExpr, Data::two_indices(cond, body), loc);

        i
    }

    fn push_if_else(
        &mut self,
        cond: Index,
        body: Index,
        other: Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        let (extra_start, _) = self.extend_extra_by_indices([body, other]);
        self.set_tag_data_source_loc(
            i,
            Tag::IfElseExpr,
            Data::index_and_extra_offset(cond, extra_start),
            loc,
        );

        i
    }

    fn push_call_expr(&mut self, lhs: Index, args: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::CallExpr, Data::two_indices(lhs, args), loc);

        i
    }

    fn push_decl_ref_unresolved(
        &mut self,
        scope: Index,
        ident: intern::Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(
            i,
            Tag::DeclRefUnresolved,
            Data::index_and_intern(scope, ident),
            loc,
        );

        i
    }

    fn resolve_decl_ref(&mut self, i: Index, decl: Index, decl_kind: DeclKind) {
        self.tags[i] = Tag::DeclRef;
        self.datas[i] = Data::index_and_opaque(decl, decl_kind as u32);
    }

    fn push_struct_decl<I: IntoIterator<Item = (intern::Index, Index)>>(
        &mut self,
        name: intern::Index,
        flags: StructFlags,
        fields: I,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node(Kind::GlobalDecl);
        self.set_struct_decl(i, name, flags, fields, loc)
    }

    fn set_struct_decl<I: IntoIterator<Item = (intern::Index, Index)>>(
        &mut self,
        i: Index,
        name: intern::Index,
        flags: StructFlags,
        fields: I,
        loc: SourceLocation,
    ) -> Index {
        let (offset, _) = self.extend_extra([flags.pack()]);
        let (names, types) = fields
            .into_iter()
            .map(|(name, ty)| (name.into_u32(), ty.into_u32()))
            .unzip::<_, _, Vec<_>, Vec<_>>();
        self.extend_extra(types);
        self.extend_extra(names);
        self.set_tag_data_source_loc(
            i,
            Tag::StructDecl,
            Data::intern_and_extra_offset(name, offset),
            loc,
        );
        i
    }

    fn push_field_decl(&mut self, name: intern::Index, ty: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::FieldDecl, Data::index_and_intern(ty, name), loc);

        i
    }

    fn push_field_access(
        &mut self,
        expr: Index,
        name: intern::Index,
        loc: SourceLocation,
    ) -> PlaceOrValue {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::FieldAccess, Data::index_and_intern(expr, name), loc);

        PlaceOrValue::Place(i)
    }

    fn push_interend_type(&mut self, ty: intern::Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::InternedType, Data::intern(ty), loc);

        i
    }

    fn push_array_type(
        &mut self,
        length_expr: Index,
        pointer_ty: Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(
            i,
            Tag::ArrayType,
            Data::two_indices(length_expr, pointer_ty),
            loc,
        );

        i
    }

    fn push_pointer_type(&mut self, ty: Index, flags: PointerFlags, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(
            i,
            Tag::PointerType,
            Data::index_and_extra_offset(ty, flags.pack() as u32),
            loc,
        );

        i
    }

    fn push_type_ref_unresolved(
        &mut self,
        scope: Index,
        ident: intern::Index,
        loc: SourceLocation,
    ) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(
            i,
            Tag::TypeDeclRefUnresolved,
            Data::index_and_intern(scope, ident),
            loc,
        );

        i
    }

    fn resolve_type_ref(&mut self, i: Index, decl: Index) {
        self.tags[i] = Tag::TypeDeclRef;
        self.datas[i] = Data::index(decl);
    }

    fn push_expr_stmt(&mut self, expr: Index) -> Index {
        let i = self.reserve_node_other();
        let loc = self.get_loc(expr);
        self.set_tag_data_source_loc(i, Tag::ExprStmt, Data::index(expr), loc);

        i
    }

    fn push_constant(&mut self, value: intern::Index, ty: Index, loc: SourceLocation) -> Index {
        let i = self.reserve_node_other();
        self.set_tag_data_source_loc(i, Tag::Constant, Data::index_and_intern(ty, value), loc);

        i
    }

    fn extend_extra_by_indices<I: IntoIterator<Item = Index>>(&mut self, indices: I) -> (u32, u32) {
        self.extend_extra(indices.into_iter().map(|i| i.as_u32()))
    }
    fn extend_extra<I: IntoIterator<Item = u32>>(&mut self, words: I) -> (u32, u32) {
        let i = self.extra.len() as u32;
        self.extra.extend(words);

        (i, self.extra.len() as u32)
    }
    fn set_tag_data_source_loc(&mut self, index: Index, tag: Tag, data: Data, loc: SourceLocation) {
        self.tags[index] = tag;
        self.datas[index] = data;
        self.source_locs[index] = loc;
    }
}

struct Children(Vec<Index>);

impl Display for Children {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "[")?;
        if let Some((last, rest)) = self.0.split_last() {
            for i in rest {
                write!(f, "{i}, ")?;
            }
            write!(f, "{last}")?;
        }
        write!(f, "]")
    }
}

type TypeCache = BTreeMap<Index, intern::Index>;
// type ComptimeCache = BTreeMap<Index, bool>;

#[derive(Debug, Default)]
struct ComptimeCache {
    inner: BTreeMap<Index, bool>,
    // this is (a,b) where b dominates a
    // meaning:
    //  when a is marked as runtime, b can be updated to be runtime as well.
    dependencies: BTreeMap<Index, Index>,
}

impl ComptimeCache {
    fn get(&self, key: &Index) -> Option<bool> {
        self.inner.get(key).cloned()
    }
    fn insert(&mut self, key: Index, value: bool) {
        self.inner.insert(key, value);
        if !value {
            self.set_runtime(key);
        }
    }
    fn mark_as_dominated_by(&mut self, a: Index, b: Index) {
        self.dependencies.insert(a, b);
    }
    fn set_runtime(&mut self, key: Index) {
        self.inner.insert(key, false);
        if let Some(&dom) = self.dependencies.get(&key) {
            self.set_runtime(dom);
        }
    }
}

impl Ast {
    fn get_type_of_node(
        &self,
        ip: &InternPool,
        cache: &mut TypeCache,
        index: Index,
    ) -> intern::Index {
        if let Some(ty) = cache.get(&index) {
            return *ty;
        }

        let void = ip.get_void_type();
        let tag = self.tags[index];
        let data = self.datas[index];

        let ty = match tag {
            Tag::ArgumentList
            | Tag::ExprStmt
            | Tag::ReturnExprStmt
            | Tag::Block
            | Tag::ParameterList
            | Tag::File => void,
            // these all evaluate to pointers
            Tag::VarDecl | Tag::MutVarDecl | Tag::VarDeclAssignment | Tag::MutVarDeclAssignment => {
                let (_, b) = data.as_extra_range();
                let pointee =
                    self.get_type_of_node(ip, cache, Index::from_u32(self.extra[b - 1]).unwrap());

                ip.as_mut().get_pointer_type(
                    pointee,
                    Some(PointerFlags::new(tag == Tag::VarDecl, false, false)),
                )
            }
            // these all evaluate to pointers
            Tag::GlobalDecl => {
                let (_, a) = data.as_intern_and_extra_offset();
                let pointee =
                    self.get_type_of_node(ip, cache, Index::from_u32(self.extra[a]).unwrap());

                ip.as_mut()
                    .get_pointer_type(pointee, Some(PointerFlags::new(true, false, true)))
            }
            Tag::FunctionDecl => self.get_type_of_node(ip, cache, data.as_two_indices().0),
            Tag::FunctionProto => {
                let (_, i) = data.as_intern_and_extra_offset();
                let (return_type, parameter_list) = (
                    Index::from_u32(self.extra[i]).unwrap(),
                    Index::from_u32(self.extra[i + 1]).unwrap(),
                );
                let return_type = self.datas[return_type].as_intern();
                let parameters = {
                    let (a, b) = self.datas[parameter_list].as_extra_range();
                    self.extra[a..b]
                        .iter()
                        .map(|&i| Index::from_u32(i).unwrap())
                        .map(|i| {
                            // i is index to a parameter, a parameter is (index, intern)
                            let ty = self.datas[i].as_index_intern().0;
                            self.datas[ty].as_intern()
                        })
                };

                ip.as_mut().get_function_type(return_type, parameters)
            }
            Tag::FunctionProtoInterned | Tag::StructDeclInterned => {
                let (_, ty) = data.as_two_interns();

                ty
            }
            Tag::BlockTrailingExpr => {
                let (_a, b) = data.as_extra_range();
                self.get_type_of_node(ip, cache, Index::from_u32(self.extra[b - 1]).unwrap())
            }
            Tag::CallExpr => {
                let (expr, _args) = data.as_two_indices();
                let fn_ty = self.get_type_of_node(ip, cache, expr);
                if let intern::Key::FunctionType { return_type, .. } = ip.get_key(fn_ty) {
                    return_type
                } else {
                    eprintln!(
                        "lhs of call expr is not a function: {fn_ty}: {:?}",
                        ip.get_key(fn_ty)
                    );
                    void
                }
            }
            Tag::Argument => self.get_type_of_node(ip, cache, data.as_index()),
            Tag::NamedArgument => {
                let (a, _) = data.as_index_intern();
                self.get_type_of_node(ip, cache, a)
            }
            Tag::ExplicitCast => {
                let (_, a) = data.as_two_indices();
                self.get_type_of_node(ip, cache, a)
            }
            // this evaluates to a pointer
            Tag::FieldAccess => {
                let (ty_expr, name) = data.as_index_intern();
                let ty = self.get_type_of_node(ip, cache, ty_expr);
                let pointee = match ip.get_key(ty) {
                    intern::Key::PointerType { pointee, .. }
                        if let intern::Key::StructType { fields, .. } = ip.get_key(pointee) =>
                    {
                        fields
                            .iter()
                            .cloned()
                            .find(|(n, _)| n == &name)
                            .map(|(_, t)| t)
                            .unwrap_or(void)
                    }
                    intern::Key::StructType { fields, .. } => fields
                        .iter()
                        .cloned()
                        .find(|(n, _)| n == &name)
                        .map(|(_, t)| t)
                        .unwrap_or(void),
                    _ => {
                        unimplemented!()
                    }
                };

                ip.as_mut().get_pointer_type(pointee, None)
            }
            // this evaluates to a pointer
            Tag::SubscriptExpr => {
                let ty = self.get_type_of_node(ip, cache, data.as_two_indices().0);
                match ip.get_key(ty) {
                    intern::Key::PointerType { .. } | intern::Key::ArrayType { .. } => {
                        // note: because of value-to-place and place-to-value,
                        // this is actually a pointer.
                        // pointee
                        ty
                    }
                    _ => {
                        eprintln!("lhs of subscript is not an array or pointer!");
                        void
                    }
                }
            }
            // this evaluates to a pointer
            Tag::AddressOf => {
                let ty = self.get_type_of_node(ip, cache, data.as_index());
                // TODO: find out of the expression is const, volatile for flags

                // ip.try_get_pointer_type(ty, None).unwrap()

                // note: because of value-to-place and place-to-value,
                // this is actually the same type.
                ty
            }
            Tag::Deref => {
                let ty = self.get_type_of_node(ip, cache, data.as_index());
                if let intern::Key::PointerType { pointee, .. } = ip.get_key(ty) {
                    pointee
                } else {
                    eprintln!("lhs of deref is not a pointer!");
                    void
                }
            }
            Tag::ValueToPlaceConversion => {
                let ty = self.get_type_of_node(ip, cache, data.as_index());
                ip.as_mut().get_pointer_type(ty, None)
            }
            Tag::PlaceToValueConversion => {
                let ty = self.get_type_of_node(ip, cache, data.as_index());
                ip.try_get_pointee_type(ty)
                    .unwrap_or(ip.as_mut().get_pointer_type(ty, None))
                //.expect(&format!("{} is not a pointer type", ip.display_key(ty)))
            }
            Tag::Not | Tag::Negate => self.get_type_of_node(ip, cache, data.as_index()),
            Tag::Or
            | Tag::And
            | Tag::BitOr
            | Tag::BitXOr
            | Tag::BitAnd
            | Tag::Eq
            | Tag::NEq
            | Tag::Lt
            | Tag::Gt
            | Tag::Le
            | Tag::Ge
            | Tag::Shl
            | Tag::Shr
            | Tag::Add
            | Tag::Sub
            | Tag::Mul
            | Tag::Div
            | Tag::Rem => self.get_type_of_node(ip, cache, data.as_two_indices().0),
            Tag::IfExpr => ip.get_bool_type(), // really?
            Tag::IfElseExpr => {
                let (_, b) = data.as_index_and_extra_offset();
                let if_ = Index::from_u32(self.extra[b]).unwrap();
                self.get_type_of_node(ip, cache, if_)
            }
            Tag::Constant | Tag::Parameter => {
                self.get_type_of_node(ip, cache, data.as_index_intern().0)
            }
            Tag::DeclRef => self.get_type_of_node(ip, cache, data.as_index()),
            Tag::StructDecl => {
                let (name, _) = data.as_intern_and_extra_offset();
                ip.as_mut().get_struct_type(name, index)
            }
            Tag::Assign
            | Tag::Root
            | Tag::DeclRefUnresolved
            | Tag::Error
            | Tag::Undefined
            | Tag::ReturnStmt => void,
            Tag::FieldDecl => self.get_type_of_node(ip, cache, data.as_index_intern().0),
            Tag::InternedType => data.as_intern(),
            // Tag::PointerType => {}
            // Tag::ArrayType => {}
            Tag::TypeDeclRef | Tag::TypeDeclRefUnresolved | Tag::PointerType | Tag::ArrayType => {
                ip.get_void_type()
            }
        };

        cache.insert(index, ty);
        ty
    }

    fn get_node_children(&self, index: Index) -> Vec<Index> {
        let tag = self.tags[index];
        let data = self.datas[index];

        match tag {
            Tag::File => {
                let (a, b) = data.as_extra_range();
                self.extra[a..b]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::FunctionProto => {
                let (_, i) = data.as_intern_and_extra_offset();
                self.extra[i..=i + 1]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::FunctionProtoInterned => {
                vec![]
            }
            Tag::FunctionDecl => {
                let (a, b) = data.as_two_indices();
                vec![a, b]
            }
            Tag::ParameterList => {
                let (a, b) = data.as_extra_range();
                self.extra[a..b]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::Block | Tag::BlockTrailingExpr => {
                let (a, b) = data.as_extra_range();
                self.extra[a..b]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::ExprStmt | Tag::ReturnExprStmt => {
                let a = data.as_index();
                vec![a]
            }
            Tag::VarDeclAssignment | Tag::MutVarDeclAssignment => {
                let (a, b) = data.as_extra_range();
                self.extra[a + 1..b]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::GlobalDecl => {
                let (_, offset) = data.as_intern_and_extra_offset();
                self.extra[offset..=offset + 1]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::CallExpr | Tag::ExplicitCast => {
                let (a, b) = data.as_two_indices();
                vec![a, b]
            }
            Tag::ArgumentList => {
                let (a, b) = data.as_extra_range();
                self.extra[a..b]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::Argument => {
                let a = data.as_index();
                vec![a]
            }
            Tag::FieldDecl | Tag::FieldAccess | Tag::NamedArgument => {
                let (a, _) = data.as_index_intern();
                vec![a]
            }
            Tag::ValueToPlaceConversion
            | Tag::PlaceToValueConversion
            | Tag::Deref
            | Tag::AddressOf
            | Tag::Not
            | Tag::Negate => {
                let a = data.as_index();
                vec![a]
            }
            Tag::Or
            | Tag::And
            | Tag::BitOr
            | Tag::BitXOr
            | Tag::BitAnd
            | Tag::Eq
            | Tag::NEq
            | Tag::Lt
            | Tag::Gt
            | Tag::Le
            | Tag::Ge
            | Tag::Shl
            | Tag::Shr
            | Tag::Add
            | Tag::Sub
            | Tag::Mul
            | Tag::Div
            | Tag::Rem
            | Tag::Assign
            | Tag::SubscriptExpr
            | Tag::ArrayType
            | Tag::IfExpr => {
                let (a, b) = data.as_two_indices();
                vec![a, b]
            }
            Tag::IfElseExpr => {
                let (a, b) = data.as_index_and_extra_offset();
                let if_ = Index::from_u32(self.extra[b]).unwrap();
                let else_ = Index::from_u32(self.extra[b + 1]).unwrap();
                vec![a, if_, else_]
            }
            Tag::PointerType => {
                let (a, _) = data.as_index_and_extra_offset();
                vec![a]
            }
            Tag::StructDeclInterned => vec![],
            Tag::StructDecl => {
                let (_a, offset) = data.as_intern_and_extra_offset();
                let flags = StructFlags::unpack(self.extra[offset]);
                self.extra[offset + 1..(offset + 1 + flags.num_fields as usize)]
                    .iter()
                    .map(|&i| Index::from_u32(i).unwrap())
                    .collect()
            }
            Tag::InternedType
            | Tag::Root
            | Tag::TypeDeclRefUnresolved
            | Tag::DeclRefUnresolved
            | Tag::Error
            | Tag::Undefined
            | Tag::TypeDeclRef
            | Tag::DeclRef
            | Tag::ReturnStmt => vec![],
            Tag::Parameter | Tag::Constant => {
                let (a, _) = data.as_index_intern();
                vec![a]
            }
            Tag::VarDecl | Tag::MutVarDecl => {
                let (a, _) = data.as_extra_range();

                vec![Index::from_u32(self.extra[a + 1]).unwrap()]
            }
        }
    }

    fn is_node_comptime_evaluable(&self, cache: &mut ComptimeCache, index: Index) -> bool {
        if let Some(a) = cache.get(&index) {
            a
        } else {
            let tag = self.tags[index];
            let data = self.datas[index];

            let children = self.get_node_children(index);

            let are_children_comptime = |this: &Self, cache: &mut ComptimeCache| {
                children
                    .iter()
                    .all(|&i| this.is_node_comptime_evaluable(cache, i))
            };

            let is_comptime = match tag {
                Tag::Parameter => false,
                // TODO: figure out if there are function protos that arent const
                Tag::FunctionProto | Tag::FunctionProtoInterned| Tag::FunctionDecl | Tag::Block | Tag::BlockTrailingExpr => {
                    //are_children_comptime(self, cache)
                    false
                }
                Tag::Constant => true,
                Tag::ReturnStmt => true,
                Tag::ReturnExprStmt => are_children_comptime(self, cache),
                Tag::MutVarDecl | Tag::MutVarDeclAssignment => false,
                Tag::VarDecl => false,
                Tag::ValueToPlaceConversion => false,
                Tag::PlaceToValueConversion | Tag::VarDeclAssignment => {
                    are_children_comptime(self, cache)
                }
                Tag::GlobalDecl => true,
                Tag::StructDecl => true,
                Tag::FieldDecl => true,
                Tag::DeclRef => self.tags[data.as_index()] == Tag::GlobalDecl,
                Tag::InternedType | Tag::PointerType | Tag::ArrayType | Tag::TypeDeclRef => true,
                Tag::CallExpr => false, // are_children_comptime(self, cache),
                Tag::FieldAccess => {
                //     let parent = data.as_index_intern().0;
                //     cache.mark_as_dominated_by(index, parent);
                    //     self.is_node_comptime_evaluable(cache, parent)
                    false
                }
                Tag::ArgumentList => are_children_comptime(self, cache),
                Tag::Argument => self.is_node_comptime_evaluable(cache, data.as_index()),
                Tag::NamedArgument => {
                    self.is_node_comptime_evaluable(cache, data.as_index_intern().0)
                }
                Tag::ExplicitCast => {
                    self.is_node_comptime_evaluable(cache, data.as_two_indices().0)
                }
                Tag::Deref | Tag::AddressOf => false,
                Tag::Not
                | Tag::Negate
                | Tag::Or
                | Tag::And
                | Tag::BitOr
                | Tag::BitXOr
                | Tag::BitAnd
                | Tag::Eq
                | Tag::NEq
                | Tag::Lt
                | Tag::Gt
                | Tag::Le
                | Tag::Ge
                | Tag::Shl
                | Tag::Shr
                | Tag::Add
                | Tag::Sub
                | Tag::Mul
                | Tag::Div
                    | Tag::SubscriptExpr // TODO: add array decl expression
                | Tag::Rem => are_children_comptime(self, cache),
                Tag::Assign => {
                    let (left, _) = data.as_two_indices();
                    cache.mark_as_dominated_by(index, left);
                    are_children_comptime(self, cache)
                }
                Tag::IfExpr | Tag::IfElseExpr => false, // are_children_comptime(self, cache),
                Tag::Root
                | Tag::File
                | Tag::ParameterList
                | Tag::ExprStmt
                | Tag::DeclRefUnresolved
                | Tag::TypeDeclRefUnresolved
                | Tag::Error
                | Tag::Undefined => false,
                _ => false,
            };

            cache.insert(index, is_comptime);

            is_comptime
        }
    }

    fn comptime_value_of_node(
        &self,
        ip: &InternPool,
        pointer_bits: u16,
        cache: &mut TypeCache,
        index: Index,
    ) -> Option<crate::comptime::ComptimeNumber> {
        let tag = self.tags[index];
        let data = self.datas[index];

        match tag {
            Tag::Constant => {
                let (ty, value) = data.as_index_intern();
                let ty = self.get_type_of_node(ip, cache, ty);
                Some(interned_type_and_value_to_comptime_number(
                    ip,
                    pointer_bits,
                    ty,
                    value,
                ))
            }
            Tag::GlobalDecl => {
                let (_, offset) = data.as_intern_and_extra_offset();
                self.comptime_value_of_node(
                    ip,
                    pointer_bits,
                    cache,
                    Index::from_u32(self.extra[offset + 1]).unwrap(),
                )
            }
            Tag::VarDeclAssignment => {
                let (a, _) = data.as_extra_range();
                self.comptime_value_of_node(
                    ip,
                    pointer_bits,
                    cache,
                    Index::from_u32(self.extra[a + 1]).unwrap(),
                )
            }
            Tag::SubscriptExpr => {
                todo!()
            }
            Tag::PlaceToValueConversion | Tag::DeclRef => {
                self.comptime_value_of_node(ip, pointer_bits, cache, data.as_index())
            }
            Tag::ExplicitCast => {
                let (expr, ty) = data.as_two_indices();
                let ty = ip.as_ast1_type(intern::AMD64_POINTER_BITS, self.datas[ty].as_intern());
                let val = self.comptime_value_of_node(ip, pointer_bits, cache, expr);
                val.and_then(|i| i.explicit_cast(ty).ok())
            }
            Tag::Add => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.add(b).ok())
                    })
            }
            Tag::Sub => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.sub(b).ok())
                    })
            }
            Tag::Mul => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.mul(b).ok())
                    })
            }
            Tag::Div => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.div(b).ok())
                    })
            }
            Tag::Rem => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.rem(b).ok())
                    })
            }
            Tag::Shl => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.shl(b).ok())
                    })
            }
            Tag::Shr => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.shr(b).ok())
                    })
            }
            Tag::BitAnd => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.bitand(b).ok())
                    })
            }
            Tag::BitOr => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.bitor(b).ok())
                    })
            }
            Tag::BitXOr => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.bitxor(b).ok())
                    })
            }
            Tag::And => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.and(b).ok())
                    })
            }
            Tag::Or => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.or(b).ok())
                    })
            }
            Tag::Eq => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.eq(b).ok())
                    })
            }
            Tag::NEq => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.eq(b).and_then(|i| i.not()).ok())
                    })
            }
            Tag::Gt => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.gt(b).ok())
                    })
            }
            Tag::Lt => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.lt(b).ok())
                    })
            }
            Tag::Le => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.le(b).ok())
                    })
            }
            Tag::Ge => {
                let (a, b) = data.as_two_indices();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| {
                        self.comptime_value_of_node(ip, pointer_bits, cache, b)
                            .and_then(|b| a.ge(b).ok())
                    })
            }
            Tag::Not => {
                let a = data.as_index();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| a.not().ok())
            }
            Tag::Negate => {
                let a = data.as_index();
                self.comptime_value_of_node(ip, pointer_bits, cache, a)
                    .and_then(|a| a.neg().ok())
            }
            _ => None,
        }
    }
}

pub fn comptime_number_to_interned_type_and_value(
    ip: &mut intern::InternPool,
    pointer_bits: u16,
    comptime: ComptimeNumber,
) -> (intern::Index, intern::Index) {
    use crate::ast::IntegralType;

    let (value, ty) = comptime.into_bytes_and_type();
    let value = match ty {
        crate::ast::Type::Bool => {
            if value.get(0) != Some(&1) {
                ip.get_false_value()
            } else {
                ip.get_true_value()
            }
        }
        crate::ast::Type::Integer(IntegralType { bits: 65.., .. })
        | crate::ast::Type::ComptimeNumber => {
            let bigint = BigInt::from_signed_bytes_le(&value);
            if bigint.sign() == num_bigint::Sign::Minus {
                ip.get_or_insert(intern::Key::NegativeInt { bigint })
            } else {
                ip.get_or_insert(intern::Key::PositiveInt { bigint })
            }
        }
        crate::ast::Type::Integer(i) => match i.bits {
            ..=32 => {
                let mut buf = [0u8; 4];
                buf[..value.len()].copy_from_slice(&value[..]);
                if i.signed {
                    ip.get_or_insert(intern::Key::SIntSmall {
                        bits: i32::from_le_bytes(buf),
                    })
                } else {
                    ip.get_or_insert(intern::Key::UIntSmall {
                        bits: u32::from_le_bytes(buf),
                    })
                }
            }
            ..=64 => {
                let mut buf = [0u8; 8];
                buf[..value.len()].copy_from_slice(&value[..]);
                if i.signed {
                    ip.get_or_insert(intern::Key::SInt64 {
                        bits: i64::from_le_bytes(buf),
                    })
                } else {
                    ip.get_or_insert(intern::Key::UInt64 {
                        bits: u64::from_le_bytes(buf),
                    })
                }
            }
            _ => unreachable!(),
        },
        crate::ast::Type::Floating(FloatingType::Binary32) => {
            let mut buf = [0u8; 4];
            buf[..value.len()].copy_from_slice(&value[..]);
            ip.get_or_insert(intern::Key::F32 {
                bits: f32::from_le_bytes(buf),
            })
        }
        crate::ast::Type::Floating(FloatingType::Binary64) => {
            let mut buf = [0u8; 8];
            buf[..value.len()].copy_from_slice(&value[..]);

            ip.get_or_insert(intern::Key::F64 {
                bits: f64::from_le_bytes(buf),
            })
        }
        _ => unimplemented!(),
    };
    let ty = ip.from_ast1_type(pointer_bits, &ty);

    (value, ty)
}

pub fn interned_type_and_value_to_comptime_number<IP>(
    ip: IP,
    pointer_bits: u16,
    ty: intern::Index,
    val: intern::Index,
) -> crate::comptime::ComptimeNumber
where
    IP: AsRef<intern::InternPool>,
{
    use crate::ast::IntegralType;
    use crate::comptime::*;

    let ip = ip.as_ref();

    let ty_key = ip.get_key(ty);
    let signed = ip.is_type_signed(ty, AMD64_POINTER_TYPE_INFO);
    match ty_key {
        intern::Key::SIntType { bit_width: bits } | intern::Key::UIntType { bit_width: bits } => {
            let ty = IntegralType::new(signed, bits);
            match ip.get_key(val) {
                intern::Key::SIntSmall { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                    bits: bits as _,
                    ty,
                }),
                intern::Key::UIntSmall { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                    bits: bits as _,
                    ty,
                }),
                intern::Key::SInt64 { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                    bits: bits as _,
                    ty,
                }),
                intern::Key::UInt64 { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                    bits: bits as _,
                    ty,
                }),
                intern::Key::PositiveInt { bigint } => {
                    ComptimeNumber::Integral(ComptimeInt::BigInt { bits: bigint, ty })
                }
                intern::Key::NegativeInt { bigint } => {
                    ComptimeNumber::Integral(ComptimeInt::BigInt { bits: bigint, ty })
                }
                _ => {
                    unreachable!()
                }
            }
        }
        intern::Key::SimpleType { ty } => match ty {
            intern::SimpleType::F32 => match ip.get_key(val) {
                intern::Key::F32 { bits } => {
                    ComptimeNumber::Floating(ComptimeFloat::Binary32(bits))
                }
                _ => {
                    unreachable!()
                }
            },
            intern::SimpleType::F64 => match ip.get_key(val) {
                intern::Key::F64 { bits } => {
                    ComptimeNumber::Floating(ComptimeFloat::Binary64(bits))
                }
                _ => {
                    unreachable!()
                }
            },
            intern::SimpleType::Bool => match ip.get_key(val) {
                intern::Key::TrueValue => ComptimeNumber::Bool(true),
                intern::Key::FalseValue => ComptimeNumber::Bool(false),
                _ => unreachable!(),
            },
            intern::SimpleType::Void => todo!(),
            intern::SimpleType::USize | intern::SimpleType::ISize => {
                let ty = IntegralType::new(
                    matches!(
                        ty_key,
                        intern::Key::SimpleType {
                            ty: intern::SimpleType::ISize
                        }
                    ),
                    pointer_bits,
                );

                match ip.get_key(val) {
                    intern::Key::SIntSmall { bits } => {
                        ComptimeNumber::Integral(ComptimeInt::Native {
                            bits: bits as _,
                            ty,
                        })
                    }
                    intern::Key::UIntSmall { bits } => {
                        ComptimeNumber::Integral(ComptimeInt::Native {
                            bits: bits as _,
                            ty,
                        })
                    }
                    intern::Key::SInt64 { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                        bits: bits as _,
                        ty,
                    }),
                    intern::Key::UInt64 { bits } => ComptimeNumber::Integral(ComptimeInt::Native {
                        bits: bits as _,
                        ty,
                    }),
                    intern::Key::PositiveInt { bigint } => {
                        ComptimeNumber::Integral(ComptimeInt::BigInt { bits: bigint, ty })
                    }
                    intern::Key::NegativeInt { bigint } => {
                        ComptimeNumber::Integral(ComptimeInt::BigInt { bits: bigint, ty })
                    }
                    _ => {
                        unreachable!()
                    }
                }
            }
            intern::SimpleType::ComptimeInt => {
                let bigint = match ip.get_key(val) {
                    intern::Key::SIntSmall { bits } => {
                        BigInt::from_signed_bytes_le(&bits.to_le_bytes())
                    }
                    intern::Key::UIntSmall { bits } => {
                        BigInt::from_signed_bytes_le(&bits.to_le_bytes())
                    }
                    intern::Key::SInt64 { bits } => {
                        BigInt::from_signed_bytes_le(&bits.to_le_bytes())
                    }
                    intern::Key::UInt64 { bits } => {
                        BigInt::from_signed_bytes_le(&bits.to_le_bytes())
                    }
                    intern::Key::PositiveInt { bigint } | intern::Key::NegativeInt { bigint } => {
                        bigint
                    }
                    _ => {
                        unreachable!()
                    }
                };
                ComptimeNumber::Integral(ComptimeInt::Comptime(bigint))
            }
        },
        _ => {
            unreachable!()
        }
    }
}
